package fr.bhprod.sga;

import java.io.File;
import java.util.logging.Level;

import com.jme.app.SimpleGame;
import com.jme.image.Texture;
import com.jme.input.AbsoluteMouse;
import com.jme.input.KeyBindingManager;
import com.jme.input.KeyInput;
import com.jme.input.MouseInput;
import com.jme.math.Vector3f;
import com.jme.scene.Node;
import com.jme.scene.state.AlphaState;
import com.jme.scene.state.TextureState;
import com.jme.util.LoggingSystem;
import com.jme.util.TextureManager;
import com.jme.util.Timer;

import fr.bhprod.sga.config.GameState;
import fr.bhprod.sga.entity.Crate;
import fr.bhprod.sga.entity.PlayableObject;
import fr.bhprod.sga.entity.UnitObject;
import fr.bhprod.sga.gui.KeyManager;
import fr.bhprod.sga.gui.MouseManager;
import fr.bhprod.sga.terrain.Terrain;
import fr.bhprod.sga.terrain.TerrainManager;

public class Sga extends SimpleGame {
	private AbsoluteMouse mouse = null;
	private UnitObject soldier = null;
	private PlayableObject crate = null;
	private Terrain terrain = null;
	private GameState state = null;
	private KeyManager keyManager = null;
	private boolean isFirstTime = true;

	public static void main(String[] args) {
		Sga app = new Sga();
		app.setDialogBehaviour(FIRSTRUN_OR_NOCONFIGFILE_SHOW_PROPS_DIALOG);
		app.start();
	}
	
	@Override
	protected void simpleInitGame() {
        LoggingSystem.getLogger().setLevel(Level.OFF);
        
		state = new GameState(timer);
		
//		input = new InputHandler();
		
		display.setTitle("SGA");
        
        /*CullState cs = display.getRenderer().createCullState();
        cs.setCullMode(CullState.CS_BACK);
        cs.setEnabled(true);
        rootNode.setRenderState(cs);

        ZBufferState buf = display.getRenderer().createZBufferState();
        buf.setFunction(ZBufferState.CF_LEQUAL);
        buf.setEnabled(true);
        rootNode.setRenderState(buf);*/
        
        // Create the terrain node
        TerrainManager terrainManager = new TerrainManager();
        terrain = terrainManager.load(new File("media/maps/bridge.map"), display.getRenderer());
		rootNode.attachChild(terrain.getNode());
		
		// Create the units node
		Node unitsNode = new Node("units");
		
		try {
			soldier = new UnitObject("soldier", new Vector3f(0,100,0), new File("media/models/soldat.xml"), terrain, display.getRenderer());
			soldier.setAnimationCycle(false);
			soldier.setAnimation("stand");
			unitsNode.attachChild(soldier.getNode());
		} catch (Throwable e) {
			e.printStackTrace();
			System.exit(1);
		}
		state.getUnits().put(soldier.getNode().getName(), soldier);
		
		rootNode.attachChild(unitsNode);

		// Create the objects node
		Node objectsNode = new Node("objects");
		
		try {
			crate = new Crate("crate01", new Vector3f(2,20,-2), display.getRenderer());
			objectsNode.attachChild(crate.getNode());
		} catch (Throwable e) {
			e.printStackTrace();
			System.exit(1);
		}
		state.getObjects().put(crate.getNode().getName(), crate);

		rootNode.attachChild(objectsNode);
		
		// Put the camera on position
		
		/*cam.setLocation(new Vector3f(0,50,-100));
		cam.lookAt(new Vector3f(68,20,-132), Vector3f.UNIT_Y);*/
		cam.setLocation(new Vector3f(36.431885f, 29.762047f, -199.84944f));
		cam.lookAt(new Vector3f(36.431885f+0.79854107f,29.762047f-0.27616656f,-199.84944f+0.5348498f), new Vector3f(0.22945377f, 0.9611098f, 0.15368453f));
		cam.update();
		
		lightState.setEnabled(false);
		

		
		mouse = new AbsoluteMouse("Mouse Input", display.getWidth(), display.getHeight());
        mouse.setLocalTranslation(new Vector3f(display.getWidth()/2, display.getHeight()/2,0));
        TextureState cursor = display.getRenderer().createTextureState();
        cursor.setEnabled(true);
        cursor.setTexture(
        			TextureManager.loadTexture(
        					Sga.class.getClassLoader().getResource("jmetest/data/cursor/cursor1.png"),
							Texture.MM_LINEAR, Texture.FM_LINEAR)
					);
        mouse.setRenderState(cursor);
        mouse.registerWithInputHandler( input );

        AlphaState as1 = display.getRenderer().createAlphaState();
        as1.setBlendEnabled(true);
        as1.setSrcFunction(AlphaState.SB_ONE);
        as1.setDstFunction(AlphaState.DB_ONE_MINUS_SRC_COLOR);
        as1.setTestEnabled(true);
        as1.setTestFunction(AlphaState.TF_GREATER);
        mouse.setRenderState(as1);
        rootNode.attachChild(mouse);

        MouseInput.get().addListener(new MouseManager(rootNode, cam, display, state, terrain));
        
		
        
        KeyBindingManager.getKeyBindingManager().add("alt", KeyInput.KEY_LMENU);
		KeyBindingManager.getKeyBindingManager().add("alt", KeyInput.KEY_RMENU);
		KeyBindingManager.getKeyBindingManager().add("cam", KeyInput.KEY_C);
		
		keyManager = new KeyManager(state);
        
        // IMPORTANT
		rootNode.updateRenderState();
	}

	@Override
	protected void simpleRender() {
		super.simpleRender();
	}

	@Override
	protected void simpleUpdate() {
		super.simpleUpdate();
		
/*		for (Iterator<IMovable> it = state.getMovableObjects().iterator(); it.hasNext();) {
			it.next().move();
		}*/
		
		keyManager.doAction(cam);
		
		if (isFirstTime) {
			soldier.putOnTerrain(new Vector3f(68,100,-132), terrain, display.getRenderer());
soldier.goToDest(new Vector3f(98,100,-192), timer.getTime(), terrain, rootNode);
			isFirstTime = false;
		}
		
		soldier.move(timer.getTime(), rootNode);
	}
}
