package fr.bhprod.sga.entity;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.xml.sax.SAXException;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public class XmlModelHandler {

	private XmlModelHandler() {
	}
	
	public static XmlModel load(File xmlModelFile) {
		SAXReader xmlReader = new SAXReader(true);
		try {
			xmlReader.setFeature("http://apache.org/xml/features/validation/schema", true);
		} catch (SAXException e) {
			e.printStackTrace();
			System.exit(1);
		}
	    
		Document doc = null;
		try {
			doc = xmlReader.read(xmlModelFile);
		} catch (DocumentException e) {
			e.printStackTrace();
			System.exit(1);
		}

	    XmlModel model = new XmlModel();
	    model.setFile(doc.valueOf("/model/model-file"));
	    model.setTextureFile(doc.valueOf("/model/texture-file"));
	    model.setSpeed(Float.parseFloat(doc.valueOf("/model/speed")));
	    model.setScale(Float.parseFloat(doc.valueOf("/model/scale")));
	    List animationsNodes = doc.selectNodes("/model/animation");
	    HashMap<String,Animation> animations = null;
	    if (!animationsNodes.isEmpty()) {
		    animations = new HashMap<String,Animation>();
		    for (Iterator iter = animationsNodes.iterator(); iter.hasNext();) {
		        Element element = (Element) iter.next();
		        Animation animation = new Animation();
		        animation.setName(element.valueOf("./name"));
		        animation.setStartFrame(Float.parseFloat(element.valueOf("./start-frame")));
		        animation.setEndFrame(Float.parseFloat(element.valueOf("./end-frame")));
		        animations.put(animation.getName(), animation);
		    }
		    model.setAnimations(animations);
		    model.setAnimated(true);
	    } else {
		    model.setAnimated(false);
	    }

	    return model;
	}

}

