package fr.bhprod.sga.gui;

import java.util.ArrayList;
import java.util.Enumeration;

import org.apache.log4j.Logger;

import com.jme.input.MouseInputListener;
import com.jme.intersection.PickResults;
import com.jme.intersection.TrianglePickResults;
import com.jme.math.Matrix3f;
import com.jme.math.Ray;
import com.jme.math.Triangle;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.Camera;
import com.jme.scene.Node;
import com.jme.scene.Spatial;
import com.jme.scene.TriMesh;
import com.jme.system.DisplaySystem;

import fr.bhprod.sga.config.GameState;
import fr.bhprod.sga.terrain.Terrain;
import fr.bhprod.sga.terrain.TerrainPosition;
import fr.bhprod.sga.terrain.TerrainTools;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public class MouseManager implements MouseInputListener {
	private static final float MOUSE_CAMERA_TURN_SPEED = -0.01f;
	private static final float MOUSE_CAMERA_ZOOM_SPEED = 0.05f;
	private Node rootNode = null;
	private PickResults results = null;
	private DisplaySystem display = null;
	private GameState state = null;
	private Terrain terrain = null;
	private Camera camera = null;
	public static final Logger LOGGER = Logger.getLogger(MouseManager.class);

	public MouseManager(Node pRootNode, Camera pCam, DisplaySystem pDisplay, GameState pState, Terrain pTerrain) {
		super();
		rootNode = pRootNode;
		display = pDisplay;
		state = pState;
		terrain = pTerrain;
		camera = pCam;
		results = new TrianglePickResults();
        //results.setCheckDistance(true);
	}

	public void onButton(int button, boolean pressed, int x, int y) {
		System.err.println("BUTTON");
		if (pressed) {
			if (button == 0) {
				// select an object
				clickLeft(x, y);
				state.setClickLeft(true);
			} else if (button == 1) {
				// do an action
				clickRight(x, y);
				state.setClickRight(true);
			}
		} else {
			if (button == 0) {
				state.setClickLeft(false);
			} else if (button == 1) {
				state.setClickRight(false);
			}
		}
	}

	public void onWheel(int wheelDelta, int x, int y) {
		// TODO : Mettre des bornes
        Vector3f loc = camera.getLocation();
        Vector3f tempVa = new Vector3f();
        if ( !camera.isParallelProjection() ) {
            loc.addLocal(camera.getDirection().mult(wheelDelta * MOUSE_CAMERA_ZOOM_SPEED, tempVa));
        } else {
            // move up instead of forward if in parallel mode
            loc.addLocal(camera.getUp().mult(wheelDelta * MOUSE_CAMERA_ZOOM_SPEED, tempVa));
        }
        camera.update();
	}

	public void onMove(int xDelta, int yDelta, int newX, int newY) {
		// TODO : La camra doit suivre le relief
		if (state.getKeys().contains("ALT")) {
			Matrix3f incr = new Matrix3f();
			if (xDelta != 0) {
	            incr.fromAngleNormalAxis(xDelta * MOUSE_CAMERA_TURN_SPEED, camera.getUp());
		        incr.mult(camera.getUp(), camera.getUp());
		        incr.mult(camera.getLeft(), camera.getLeft());
		        incr.mult(camera.getDirection(), camera.getDirection());
		        camera.normalize();
		        camera.update();
			}
			// TODO : La camra doit tourner autour du point de pivot
			if (yDelta != 0) {
				incr.fromAngleNormalAxis(yDelta * MOUSE_CAMERA_TURN_SPEED, camera.getLeft());
		        incr.mult(camera.getLeft(), camera.getLeft());
		        incr.mult(camera.getDirection(), camera.getDirection());
		        incr.mult(camera.getUp(), camera.getUp());
		        camera.normalize();
		        camera.update();
			}
		}
	}

	private void clickLeft(int x, int y) {
		// Unselect the objects
		for (Enumeration<String> en = state.getSelectedObjects().keys(); en.hasMoreElements();) {
			String objectName = en.nextElement();
			if (state.getSelectedObjects().get(objectName).equals(GameState.TYPE_UNIT)) {
	    		state.getUnits().get(objectName).destroySelectionShape();
			} else if (state.getSelectedObjects().get(objectName).equals(GameState.TYPE_OBJECT)) {
	    		state.getObjects().get(objectName).destroySelectionShape();
			}
		}
		state.getSelectedObjects().clear();
		
        Vector2f screenPos = new Vector2f(x, y);
		Vector3f oriWorldCoords = display.getWorldCoordinates(screenPos, 0);
		Vector3f destWorldCoords = display.getWorldCoordinates(screenPos, 1);

        Ray ray = new Ray(oriWorldCoords, destWorldCoords.subtract(oriWorldCoords));

        // Test if the player selected a unit
        MouseSelection ms = selectEntity(ray, rootNode.getChild("units"));
        if (ms != null && ms.getMesh() != null) {
    		state.getSelectedObjects().clear();
    		state.getSelectedObjects().put(ms.getMesh().getName(),GameState.TYPE_UNIT);
    		state.getUnits().get(ms.getMesh().getName()).createSelectionShape(terrain, rootNode);
        } else {
            // Test if the player selected an object
        	ms = selectEntity(ray, rootNode.getChild("objects"));
            if (ms != null && ms.getMesh() != null) {
        		state.getSelectedObjects().clear();
        		state.getSelectedObjects().put(ms.getMesh().getName(),GameState.TYPE_OBJECT);
        		state.getObjects().get(ms.getMesh().getName()).createSelectionShape(terrain, rootNode);
            }
        }
	}

	private void clickRight(int x, int y) {
		// Do nothing if there is no selected objects
		if (state.getSelectedObjects().isEmpty()) {
			return;
		}
		
        Vector2f screenPos = new Vector2f(x, y);
		Vector3f oriWorldCoords = display.getWorldCoordinates(screenPos, 0);
		Vector3f destWorldCoords = display.getWorldCoordinates(screenPos, 1);

        Ray ray = new Ray(oriWorldCoords, destWorldCoords.subtract(oriWorldCoords));

        // Test if the player selected a unit
        MouseSelection ms = selectEntity(ray, rootNode.getChild("units"));
        if (ms != null && ms.getMesh() != null) {
        	// TODO : A finir : attaque en v1, action en v2
        	return;
        }
		
        // Test if the player selected an object
        ms = selectEntity(ray, rootNode.getChild("objects"));
        if (ms != null && ms.getMesh() != null) {
        	// TODO : A finir en v2
        	return;
        }

        // Test if the player selected the terrain
        ms = selectEntity(ray, rootNode.getChild("layerTextureStampingNode"));
        if (ms != null && ms.getMesh() != null) {
    		for (Enumeration<String> en = state.getSelectedObjects().keys(); en.hasMoreElements();) {
    			String objectName = en.nextElement();
    			// Move only if it's a unit
    			if (state.getSelectedObjects().get(objectName).equals(GameState.TYPE_UNIT)) {
    				state.getUnits().get(objectName).setDestination(ms.getClickPosition(), state.getTimer().getTime(), rootNode);
    				state.getMovableObjects().add(state.getUnits().get(objectName));
    			}
    		}
        	return;
        }
	}
	
	private MouseSelection selectEntity(Ray ray, Spatial node) {
        results.clear();

        node.findPick(ray,results);

		if(results.getNumber() > 0) {
        	MouseSelection ms = new MouseSelection();
        	ms.setMesh((TriMesh)results.getPickData(0).getTargetMesh().getParentGeom());
        	LOGGER.debug("MESH NAME : "+ms.getMesh().getName());
        	
        	ArrayList tris = results.getPickData(0).getTargetTris();
        	if (tris != null && !tris.isEmpty()) {
        		Vector3f[] vertices = new Vector3f[3];
        		ms.getMesh().getTriangle(((Integer)tris.get(0)).intValue(), vertices);
        		TerrainPosition tp = TerrainTools.getClickCoordinates(ray, new Triangle(vertices[0], vertices[1], vertices[2]));
        		if (tp != null) {
        			ms.setClickPosition(tp.getPosition());
        		}
        	}
        	
			return ms;
        } else {
        	return null;
        }
	}

}

