package fr.bhprod.sga.map.editor.gl;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.File;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import com.jme.bounding.BoundingBox;
import com.jme.image.Texture;
import com.jme.input.action.InputActionEvent;
import com.jme.input.action.KeyBackwardAction;
import com.jme.input.action.KeyForwardAction;
import com.jme.input.action.KeyLookDownAction;
import com.jme.input.action.KeyLookUpAction;
import com.jme.input.action.KeyRotateLeftAction;
import com.jme.input.action.KeyRotateRightAction;
import com.jme.input.action.KeyStrafeLeftAction;
import com.jme.input.action.KeyStrafeRightAction;
import com.jme.intersection.BoundingPickResults;
import com.jme.intersection.PickResults;
import com.jme.math.Ray;
import com.jme.math.Triangle;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.scene.TriMesh;
import com.jme.scene.shape.Box;
import com.jme.scene.state.CullState;
import com.jme.scene.state.RenderState;
import com.jme.scene.state.TextureState;
import com.jme.scene.state.WireframeState;
import com.jme.scene.state.ZBufferState;
import com.jme.util.TextureManager;
import com.jme.util.geom.Debugger;
import com.jmex.awt.SimpleCanvasImpl;

import fr.bhprod.sga.map.editor.swing.MapMenu;
import fr.bhprod.sga.terrain.Terrain;
import fr.bhprod.sga.terrain.TerrainManager;
import fr.bhprod.sga.terrain.TerrainPosition;
import fr.bhprod.sga.terrain.TerrainTools;
import fr.bhprod.sga.terrain.layer.DecalUtil;
import fr.bhprod.sga.terrain.object.Decal;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public class MapCanvasImpl extends SimpleCanvasImpl implements MouseListener, MouseMotionListener, KeyListener {

	private MapMenu menu = null;

	private boolean mouseClicked = false;

	private Vector2f mousePoint = null;

	private PickResults pr = null;

	private Terrain terrain = null;

	private Vector2f oldMousePoint = null;

	private Vector2f newMousePoint = null;

	private KeyLookDownAction lookDown = null;

	private KeyLookUpAction lookUp = null;

	private KeyRotateLeftAction rotateLeft = null;

	private KeyRotateRightAction rotateRight = null;
	
    private KeyForwardAction forward = null;

	private KeyBackwardAction backward = null;

	private KeyStrafeLeftAction strafeLeft = null;

	private KeyStrafeRightAction strafeRight = null;
	
	private String keyPressed = null;

	private static final float MOUSE_CAMERA_SPEED = 1;

	private static final float KEY_CAMERA_SPEED = 50;
	
	private boolean controlPressed = false;

	private boolean shiftPressed = false;

	private boolean leftButtonPressed = false;

	private boolean rightButtonPressed = false;
	
	private String selectedDecal = null;
	
	private boolean deleteSelectedDecal = false;
	
	private boolean drawBounds = false;
	
	private boolean firstTime = true;
	
	private File loadTerrain = null;
	
	public static final Logger LOGGER = Logger.getLogger(MapCanvasImpl.class);


	
	
	public MapCanvasImpl(int width, int height, MapMenu menuOrig) {
		super(width, height);
		menu = menuOrig;
	}
	
	@Override
	public void simpleSetup() {
		try {
			super.simpleSetup();
			
			CullState cs = renderer.createCullState();
			cs.setCullMode(CullState.CS_BACK);
			cs.setEnabled(true);
			rootNode.setRenderState(cs);

			ZBufferState buf = renderer.createZBufferState();
			buf.setFunction(ZBufferState.CF_LEQUAL);
			buf.setEnabled(true);
			rootNode.setRenderState(buf);
			
//			terrain = new Terrain(renderer);
TerrainManager loader = new TerrainManager();
terrain = loader.load(new File("media/maps/bridge.map"),renderer);
			rootNode.attachChild(terrain.getNode());
	
			/*Crate crate = new Crate("crate0", new Vector3f(0,100,0), renderer);
			crate.putOnTerrain(terrain, renderer);
			rootNode.attachChild(crate);*/
			
			cam.setLocation(new Vector3f(50, 50, -50));
			cam.lookAt(new Vector3f(0, 0, 0), Vector3f.UNIT_Y);
			cam.update();
	
			//lightState.setEnabled(false);
	
//			FogState fs = display.getRenderer().createFogState();
//			 fs.setStart(20);
//			 fs.setEnd(100);
//			 fs.setColor(ColorRGBA.orange);
//			 fs.setDensity(0.5f);
//			 rootNode.setRenderState(fs);
	
			pr = new BoundingPickResults();
			
		    lookDown = new KeyLookDownAction(cam, MOUSE_CAMERA_SPEED);
		    lookUp = new KeyLookUpAction(cam, MOUSE_CAMERA_SPEED);
		    rotateLeft = new KeyRotateLeftAction(cam, MOUSE_CAMERA_SPEED);
		    rotateRight = new KeyRotateRightAction(cam, MOUSE_CAMERA_SPEED);
		    
		    forward = new KeyForwardAction(cam, KEY_CAMERA_SPEED);
		    backward = new KeyBackwardAction(cam, KEY_CAMERA_SPEED);
		    strafeLeft = new KeyStrafeLeftAction(cam, KEY_CAMERA_SPEED);
		    strafeRight = new KeyStrafeRightAction(cam, KEY_CAMERA_SPEED);
			
			// IMPORTANT
		    rootNode.updateRenderState();
		} catch (Throwable t) {
			t.printStackTrace();
			System.exit(1);
		}
	}

	@Override
	public void simpleUpdate() {
		super.simpleUpdate();

		// Terrain loading
		if (loadTerrain != null) {
			terrain.getNode().removeFromParent();
			TerrainManager loader = new TerrainManager();
			terrain = loader.load(loadTerrain, renderer);
			rootNode.attachChild(terrain.getNode());
			rootNode.updateRenderState();
			loadTerrain = null;
		}
		
		if (firstTime) {
			firstTime = false;
		}
		
		if (deleteSelectedDecal) {
			terrain.getDecals().remove(selectedDecal);
			terrain.getLayerDecalsNode().getChild(selectedDecal).removeFromParent();
			rootNode.getChild("decal_bbox").removeFromParent();
			selectedDecal = null;
			deleteSelectedDecal = false;
		}
		
		if (mouseClicked) {
			mouseClicked = false;
			
			Vector3f worldCoords = cam.getWorldCoordinates(mousePoint, 0);
			Vector3f worldCoords2 = cam.getWorldCoordinates(mousePoint, 1);
			
			Ray mouseRay = new Ray(worldCoords, worldCoords2.subtractLocal(worldCoords));
			pr.clear();
			
			// Search for selected decal or selected detail patch
			if (menu.getCurrentMenuType() == MapMenu.MENU_TYPE_DECAL) {
				if (controlPressed && menu.getCurrentMenuType() == MapMenu.MENU_TYPE_DECAL) {
					// Decal selection
LOGGER.debug("SELECTED");
					mouseSelectDecal(mouseRay);
				} else {
					// Decal creation
					if (menu.getSelectedDecalFile() != null) {
						mouseCreateDecal(mouseRay);
					}
				}
			} else if (menu.getCurrentMenuType() == MapMenu.MENU_TYPE_TEXTURE) {
				mouseSelectPatch(mouseRay);
			}
		}
		
		if (oldMousePoint != null && newMousePoint != null && !oldMousePoint.equals(newMousePoint)) {
			if (shiftPressed && menu.getCurrentMenuType() == MapMenu.MENU_TYPE_DECAL) {
				// Decal modification
				if (leftButtonPressed) {
					// Decal rotation
LOGGER.debug("ROTATED");
					mouseRotateDecal();
				} else if (rightButtonPressed) {
					// Decal scale
LOGGER.debug("SCALED");
					mouseScaleDecal();
				}
			} else if (menu.getCurrentMenuType() == MapMenu.MENU_TYPE_DECAL) {
				if (rightButtonPressed) {
					// Decal move
LOGGER.debug("MOVED");
					mouseMoveDecal();
				}
			} else {
				mouseMoveCamera();
			}
			oldMousePoint = new Vector2f(newMousePoint);
		}

		if (keyPressed != null) {
			keyMoveCamera();
		}
		
		if (pr != null) {
			pr.clear();
		}
	}

	public void applyDetailTexture(File textureFile, String selectedPatch) {
// TODO : Mettre  jour la liste des textures et les index
		Texture detailTexture = TextureManager.loadTexture(textureFile.getAbsolutePath(), Texture.MM_LINEAR, Texture.FM_LINEAR, 1, true);
		detailTexture.setMipmapState(Texture.MM_LINEAR_LINEAR);
		detailTexture.setWrap(Texture.WM_WRAP_S_WRAP_T);

		TriMesh mesh = (TriMesh)terrain.getLayerTextureDetailNode().getChild(selectedPatch);
		TextureState ts = (TextureState)mesh.getRenderState(RenderState.RS_TEXTURE);
		ts.setTexture(detailTexture);
		mesh.updateRenderState();
	}

	public void mouseClicked(MouseEvent event) {
		mouseClicked = true;
		
		mousePoint = new Vector2f(event.getX(), height - event.getY());
	}

	public void mousePressed(MouseEvent event) {
		oldMousePoint = new Vector2f(event.getX(), event.getY());
		
		if (event.getButton() == MouseEvent.BUTTON1) {
			leftButtonPressed = true;
		}

		if (event.getButton() == MouseEvent.BUTTON3) {
			rightButtonPressed = true;
		}
	}

	public void mouseReleased(MouseEvent event) {
		oldMousePoint = null;
		newMousePoint = null;
		leftButtonPressed = false;
		rightButtonPressed = false;
	}
	
	public void mouseEntered(MouseEvent event) {}
	public void mouseExited(MouseEvent event) {}

	public Terrain getTerrain() {
		return terrain;
	}

	public void mouseDragged(MouseEvent event) {
		newMousePoint = new Vector2f(event.getX(), event.getY());
	}

	public void mouseMoved(MouseEvent event) {}

	public void keyTyped(KeyEvent event) {}

	public void keyPressed(KeyEvent event) {
// TODO : A enlever
if (event.getKeyCode() == KeyEvent.VK_ESCAPE) {
	System.exit(0);
}

		if (event.getKeyCode() == KeyEvent.VK_Z) {
			keyPressed = "forward";
		}
		
		if (event.getKeyCode() == KeyEvent.VK_S) {
			keyPressed = "backward";
		}
		
		if (event.getKeyCode() == KeyEvent.VK_Q) {
			keyPressed = "strafeLeft";
		}

		if (event.getKeyCode() == KeyEvent.VK_D) {
			keyPressed = "strafeRight";
		}

		if (event.getKeyCode() == KeyEvent.VK_CONTROL) {
			controlPressed = true;
		}

		if (event.getKeyCode() == KeyEvent.VK_SHIFT) {
			shiftPressed = true;
		}

		if (event.getKeyCode() == KeyEvent.VK_B) {
			drawBounds = !drawBounds;
		}

		if (event.getKeyCode() == KeyEvent.VK_DELETE) {
			if (menu.getCurrentMenuType() == MapMenu.MENU_TYPE_DECAL && selectedDecal != null) {
				deleteSelectedDecal = true;
			}
		}
	}

	public void keyReleased(KeyEvent event) {
		keyPressed = null;
		controlPressed = false;
		shiftPressed = false;
	}
	
	private void mouseMoveCamera() {
		Vector2f localTranslation = new Vector2f(oldMousePoint);
		localTranslation.set(localTranslation.subtract(newMousePoint));

		InputActionEvent inputEvent = new InputActionEvent();

		if (localTranslation.x > 0) {
			inputEvent.setTime(MOUSE_CAMERA_SPEED * localTranslation.x * -1 * 0.01f);
			rotateRight.performAction(inputEvent);
		} else if (localTranslation.x < 0) {
			inputEvent.setTime(MOUSE_CAMERA_SPEED * localTranslation.x * 0.01f);
			rotateLeft.performAction(inputEvent);
		}
		
		if (localTranslation.y > 0) {
			inputEvent.setTime(MOUSE_CAMERA_SPEED * localTranslation.y * -1 * 0.01f);
			lookUp.performAction(inputEvent);
		} else if (localTranslation.y < 0) {
			inputEvent.setTime(MOUSE_CAMERA_SPEED * localTranslation.y * 0.01f);
			lookDown.performAction(inputEvent);
		}
	}

	private void keyMoveCamera() {
		InputActionEvent inputEvent = new InputActionEvent();
		inputEvent.setTime(timer.getTimePerFrame());

		if (keyPressed.equals("forward")) {
			forward.performAction(inputEvent);
		}
		
		if (keyPressed.equals("backward")) {
			backward.performAction(inputEvent);
		}

		if (keyPressed.equals("strafeLeft")) {
			strafeLeft.performAction(inputEvent);
		}

		if (keyPressed.equals("strafeRight")) {
			strafeRight.performAction(inputEvent);
		}
	}
	
	private void mouseSelectPatch(Ray mouseRay) {
		terrain.getLayerTextureDetailNode().findPick(mouseRay, pr);
		if (pr.getNumber() != 0) {
			TriMesh mesh = (TriMesh)pr.getPickData(0).getTargetMesh().getParentGeom();
			menu.setSelectedPatch(mesh.getName());
			BoundingBox box = (BoundingBox)mesh.getBatch(0).getModelBound();
			
			// Display the bounding of the selected object
			Box boundingBox = (Box)rootNode.getChild("patch_bbox");
			if (boundingBox == null) {
				boundingBox = new Box("patch_bbox", box.getCenter(), box.xExtent, box.yExtent, box.zExtent);
				WireframeState ws = renderer.createWireframeState();
				ws.setEnabled(true);
				boundingBox.setRenderState(ws);
				rootNode.attachChild(boundingBox);
				rootNode.updateRenderState();
			} else {
				boundingBox.setData(box.getCenter(), box.xExtent, box.yExtent, box.zExtent);
			}
		}
	}
	
	private void mouseCreateDecal(Ray mouseRay) {
		terrain.getLayerTextureDetailNode().findPick(mouseRay, pr);

		if (pr.getNumber() != 0) {
			TriMesh mesh = (TriMesh)pr.getPickData(0).getTargetMesh().getParentGeom();
			if (mesh != null) {
				ArrayList<Integer> picks = new ArrayList<Integer>();
				mesh.findTrianglePick(mouseRay, picks, 0);
				if (picks != null && !picks.isEmpty()) {
					Vector3f[] triCoords = new Vector3f[3];
					mesh.getTriangle(picks.get(0), triCoords);
					if (triCoords != null) {
						TerrainPosition position = TerrainTools.getClickCoordinates(mouseRay, new Triangle(triCoords[0], triCoords[1], triCoords[2]));
						LOGGER.debug(position);
						if (position != null) {
							LOGGER.debug(position.getPosition());
							String decalName = "dec"+timer.getTime();
							Decal decal = DecalUtil.createNewDecal(new Vector3f(position.getPosition().x, 0, position.getPosition().z), decalName, menu.getSelectedDecalFile().getAbsolutePath(), terrain, renderer);
							terrain.getDecals().put(decalName, decal);
							terrain.getLayerDecalsNode().updateWorldBound();
						}
					}
				}
			}
		}
	}

	private void mouseSelectDecal(Ray mouseRay) {
		terrain.getLayerDecalsNode().findPick(mouseRay, pr);

		if (pr.getNumber() != 0) {
			TriMesh mesh = (TriMesh)pr.getPickData(0).getTargetMesh().getParentGeom();
			selectedDecal = mesh.getName();
			BoundingBox box = (BoundingBox)mesh.getBatch(0).getModelBound();
			
			// Display the bounding of the selected object
			Box boundingBox = (Box)rootNode.getChild("decal_bbox");
			if (boundingBox == null) {
				boundingBox = new Box("decal_bbox", box.getCenter(), box.xExtent, box.yExtent, box.zExtent);
				WireframeState ws = renderer.createWireframeState();
				ws.setEnabled(true);
				boundingBox.setRenderState(ws);
				rootNode.attachChild(boundingBox);
				rootNode.updateRenderState();
			} else {
				boundingBox.setData(box.getCenter(), box.xExtent, box.yExtent, box.zExtent);
			}
		}
	}

	private void mouseRotateDecal() {
		Vector2f localTranslation = new Vector2f(oldMousePoint);
		localTranslation.set(localTranslation.subtract(newMousePoint));

		terrain.getDecals().get(selectedDecal).rotate(MOUSE_CAMERA_SPEED * localTranslation.x * 0.01f, terrain, renderer);
	}

	private void mouseScaleDecal() {
		Vector2f localTranslation = new Vector2f(oldMousePoint);
		localTranslation.set(localTranslation.subtract(newMousePoint));

		terrain.getDecals().get(selectedDecal).scale(MOUSE_CAMERA_SPEED * localTranslation.x * 0.1f, terrain, renderer);
	}

	private void mouseMoveDecal() {
		Vector2f localTranslation = new Vector2f(oldMousePoint);
		localTranslation.set(localTranslation.subtract(newMousePoint));

		terrain.getDecals().get(selectedDecal).move(MOUSE_CAMERA_SPEED * localTranslation.x * 0.1f, MOUSE_CAMERA_SPEED * localTranslation.y * 0.1f, terrain, renderer);
	}

	@Override
	public void simpleRender() {
		super.simpleRender();
		
		if (drawBounds) {
			Debugger.drawBounds(rootNode, renderer, true);
		}
	}

	public void setTerrain(Terrain terrain) {
		this.terrain = terrain;
	}

	public void setLoadTerrain(File loadTerrain) {
		this.loadTerrain = loadTerrain;
	}

}
