package fr.bhprod.sga.terrain;

import java.io.File;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.log4j.Logger;

import com.jme.image.Texture;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.util.TextureManager;

import fr.bhprod.sga.terrain.layer.DecalUtil;
import fr.bhprod.sga.terrain.layer.TextureDetailUtil;
import fr.bhprod.sga.terrain.layer.TextureStampingUtil;
import fr.bhprod.sga.terrain.object.Decal;

/**
 * Create the terrain node with three layers.
 * 
 * 	Lighting properties 
o	Sunlight Color 
o	Shadow Color 
o	Time of Day 
	Sky 
o	Sky Texture 
o	Sky Scale 
	Fog 
o	Fog Start Distance 
o	Fog End Distance 
o	Color 
	Misc 
o	Detail Texture Repeat 
	Water 
o	Water Set 
o	Color 
o	Alpha 
o	Anim Speed 
o	Repeat

 * 
 * @author jbeaufils
 * @version 1.0
 */
// TODO: Grer les objets sur le terrain
// TODO : Rattacher cette implmentation du terrain  l'diteur de map
// FIXME : Les dimensions de la zone de jeu relle sont divises par 2 !
public class Terrain {
	
	/** Generated ID. */
	private static final long serialVersionUID = 2458711602445295202L;

	/** Number of x tiles. */
	private int tilesX = 0;

	/** Number of z tiles. */
	private int tilesZ = 0;

	/** Real size of a tile in meters. */
	private int tileRealSize = 0;

	/** Contains the heights of the vertices. */
	private float[][] heights = null;

	private Node layerTextureStampingNode = null;
	
	private Node layerTextureDetailNode = null;
	
	private Node layerDecalsNode = null;
	
	private Node node = null;
	
	private short[][] inpassMap = null;
	
	private HashMap<String,Decal> decals = null;

	private TerrainManager manager = null;
	
	public static final Logger LOGGER = Logger.getLogger(Terrain.class);
	
	

	/**
	 * Load the heightmap and create the layers.
	 * 
	 * @param display
	 *            The current display
	 */
	public Terrain(TerrainManager loader, Renderer renderer) {
		manager = loader;
		
		this.tilesX = loader.getTilesX();
		this.tilesZ = loader.getTilesZ();
		this.tileRealSize = loader.getTileRealSize();
		this.inpassMap = loader.getInpassMap();
		decals = new HashMap<String,Decal>(loader.getDecals());
		
		node = new Node("terrain");
		
		//createWater(display);
		loadHeightMap(loader.getHeightmapFile());
		
		layerTextureStampingNode = TextureStampingUtil.paintManyLayerTextureStamping(loader.getStampingFile(), this, renderer);
		node.attachChild(layerTextureStampingNode);

		layerTextureDetailNode = TextureDetailUtil.paintManyLayerTextureDetail(this, loader.getDetailsFiles(), loader.getDetailIndexes(), true, renderer);
		node.attachChild(layerTextureDetailNode);
		
		layerDecalsNode = new Node("layerDecals");
		for (Iterator<Decal> it = loader.getDecals().values().iterator(); it.hasNext();) {
			DecalUtil.putDecal(it.next(), this, renderer);
		}
		node.attachChild(layerDecalsNode);
	}

	public Terrain(Renderer renderer) {
		this.tilesX = 128;
		this.tilesZ = 256;
		this.tileRealSize = 2;
		decals = new HashMap<String,Decal>();
		
		node = new Node("terrain");
		
		//createWater(display);
		loadHeightMap(new File("media/maps/bridge_heightmap.tga"));
		
		layerTextureStampingNode = TextureStampingUtil.paintManyLayerTextureStamping(new File("media/maps/bridge_stamp.png"), this, renderer);
		node.attachChild(layerTextureStampingNode);

		layerTextureDetailNode = new Node("layerTextureDetail");
		node.attachChild(layerTextureDetailNode);
		
		layerDecalsNode = new Node("layerDecals");
		node.attachChild(layerDecalsNode);
	}

	/**
	 * Load the heightmap file and construct the heights array.
	 */
	private final void loadHeightMap(File heightmapFile) {
		Texture heightMap = TextureManager.loadTexture(heightmapFile.getAbsolutePath(),
				Texture.MM_LINEAR, Texture.FM_LINEAR, 1, true);
		ByteBuffer buffer = heightMap.getImage().getData();
		heights = new float[getVerticesX()][getVerticesZ()];
		for (int z = 0; z < getVerticesZ(); z++) {
			for (int x = 0; x < getVerticesX(); x++) {
				heights[x][z] = (float) (buffer.get(z * getVerticesX() * 4 + x * 4)) / 4f;
			}
		}
	}


	/*private void createWater(Renderer renderer) {
		Quad waterPlane = new Quad("waterPlane", TILES_X * TILE_REAL_SIZE, TILES_Z * TILE_REAL_SIZE);
		Matrix3f mat = new Matrix3f();
		mat.fromAngleAxis(90 * FastMath.DEG_TO_RAD, new Vector3f(1,0,0));
		waterPlane.setLocalRotation(mat);
		waterPlane.setLocalTranslation(new Vector3f(TILES_X * TILE_REAL_SIZE / 2,4,-(TILES_Z * TILE_REAL_SIZE / 2)));

		TextureState ts = renderer.createTextureState();
		Texture texture = TextureManager.loadTexture("media/water/water.tga",
				Texture.MM_LINEAR_LINEAR, Texture.FM_LINEAR);
		ts.setTexture(texture);

		waterPlane.setRenderState(ts);

		AlphaState as = renderer.createAlphaState();
		as.setBlendEnabled(true);
		as.setDstFunction(AlphaState.DB_ONE_MINUS_SRC_ALPHA);
		as.setSrcFunction(AlphaState.SB_SRC_ALPHA);
		waterPlane.setRenderState(as);

		attachChild(waterPlane);
	}*/
	
	public Node getLayerDecalsNode() {
		return layerDecalsNode;
	}

	public Node getLayerTextureDetailNode() {
		return layerTextureDetailNode;
	}

	public Node getLayerTextureStampingNode() {
		return layerTextureStampingNode;
	}

	public float[][] getHeights() {
		return heights;
	}

	public void setLayerDecalsNode(Node layerDecalsNode) {
		this.layerDecalsNode = layerDecalsNode;
	}

	public Node getNode() {
		return node;
	}

	public HashMap<String, Decal> getDecals() {
		return decals;
	}

	public int getTileRealSize() {
		return tileRealSize;
	}

	public int getTilesX() {
		return tilesX;
	}

	public int getTilesZ() {
		return tilesZ;
	}

	public int getVerticesX() {
		return tilesX+1;
	}

	public int getVerticesZ() {
		return tilesZ+1;
	}

	public TerrainManager getManager() {
		return manager;
	}

	public short[][] getInpassMap() {
		return inpassMap;
	}

}
