package fr.bhprod.sga.terrain;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.imageio.ImageIO;

import org.apache.log4j.Logger;

import com.jme.image.Texture;
import com.jme.math.Vector3f;
import com.jme.renderer.Renderer;
import com.jme.util.TextureManager;
import com.jmex.effects.ProjectedTextureUtil;

import fr.bhprod.sga.terrain.object.Decal;

public class TerrainManager {
	private int tilesX = 0;

	private int tilesZ = 0;

	private int tileRealSize = 0;

	private File heightmapFile = null;

	private File stampingFile = null;
	
	private int nbDetailTextureFiles = 0;

	private File[] detailsFiles = null;

	private int[][] detailIndexes = null;
	
	private int nbDecalFiles = 0;
	
	private File[] decalsFiles = null;

	private int nbDecals = 0;
	
	private short[][] inpassMap = null;
	
	private HashMap<String,Decal> decals = new HashMap<String,Decal>();
	
	public static final Logger LOGGER = Logger.getLogger(TerrainManager.class);

	public Terrain load(File file, Renderer renderer) {
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
		
		String[] values = null;

		try {
			tilesX = Integer.parseInt(readLine(br));
			tilesZ = Integer.parseInt(readLine(br));
			tileRealSize = Integer.parseInt(readLine(br));

			heightmapFile = new File(readLine(br));
			
			stampingFile = new File(readLine(br));
			
			BufferedImage image = ImageIO.read(new File(readLine(br)));
			if (image.getWidth() != tilesX || image.getHeight() != tilesZ) {
				LOGGER.error("The inpass image and the map must have the same dimensions");
				System.exit(1);
			}
			inpassMap = new short[tilesX][tilesZ];
			for (int j = 0; j < tilesZ; j++) {
				for (int i = 0; i < tilesX; i++) {
					if ((image.getRGB(i, j) & 0x000000ff) == 0) {
						inpassMap[i][j] = 0;
					} else {
						inpassMap[i][j] = 1;
					}
				}
			}

			nbDetailTextureFiles = Integer.parseInt(readLine(br));
			detailsFiles = new File[nbDetailTextureFiles];
			for (int i = 0; i < nbDetailTextureFiles; i++) {
				detailsFiles[i] = new File(readLine(br));
			}
			detailIndexes = new int[tilesX/16][tilesZ/16];
			for (int j = 0; j < tilesZ/16; j++) {
				values = readLine(br).split(" ");
				for (int i = 0; i < tilesX/16; i++) {
					detailIndexes[i][j] = Integer.parseInt(values[i]);
				}
			}
			
			nbDecalFiles = Integer.parseInt(readLine(br));
			if (nbDecalFiles != 0) {
				decalsFiles = new File[nbDecalFiles];
				for (int i = 0; i < nbDecalFiles; i++) {
					decalsFiles[i] = new File(readLine(br));
				}
				nbDecals = Integer.parseInt(readLine(br));
				for (int i = 0; i < nbDecals; i++) {
					Decal decal = new Decal();
					decal.setName(readLine(br));
LOGGER.debug("Decal name = "+decal.getName());
					values = readLine(br).split(",");
					decal.setUp(new Vector3f(Float.parseFloat(values[0]), Float.parseFloat(values[1]), Float.parseFloat(values[2])));
LOGGER.debug("Decal up = "+decal.getUp());
					values = readLine(br).split(",");
					decal.setProjectorPos(new Vector3f(Float.parseFloat(values[0]), Float.parseFloat(values[1]), Float.parseFloat(values[2])));
LOGGER.debug("Decal projectorPos = "+decal.getProjectorPos());
					int textureIndex = Integer.parseInt(readLine(br));
					Texture texture = TextureManager.loadTexture(decalsFiles[textureIndex].getAbsolutePath(), Texture.MM_LINEAR_LINEAR, Texture.FM_LINEAR);
					ProjectedTextureUtil.setupProjectedTexture(texture, Texture.WM_BCLAMP_S_BCLAMP_T, Texture.ACF_ADD);
					decal.setTexture(texture);
					decal.setSize(Float.parseFloat(readLine(br)));
					decal.setNear(Float.parseFloat(readLine(br)));
					decal.setFar(Float.parseFloat(readLine(br)));
					decal.setFov(Float.parseFloat(readLine(br)));
					decal.setAspect(Float.parseFloat(readLine(br)));
					decals.put(decal.getName(), decal);
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		
		return (new Terrain(this, renderer));
	}
	
	private String readLine(BufferedReader br) throws IOException {
		String line = null;

		do {
			line = br.readLine();
		
			// End of file
			if (line == null) {
				throw new EOFException("End of file unexpected");
			}

			// Comment line : do nothing
			if (line.startsWith("#")) {
				continue;
			}

			// Blank line : do nothing
			if (line.trim().length() == 0) {
				continue;
			}
			
			return line;
		} while (line != null);

		throw new EOFException("End of file unexpected");
	}

	public static void save(File file, Terrain terrain) {
// TODO : Sauver les objets du terrain
		PrintWriter pr = null;
		try {
			pr = new PrintWriter(new FileOutputStream(file));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
		
		pr.println(terrain.getTilesX());
		pr.println(terrain.getTilesZ());
		pr.println(terrain.getTileRealSize());
		pr.println(getMediaRelativePath(terrain.getManager().getHeightmapFile().getAbsolutePath()));
		pr.println(getMediaRelativePath(terrain.getManager().getStampingFile().getAbsolutePath()));
		pr.println(terrain.getManager().getNbDetailTextureFiles());
		for (int i = 0; i < terrain.getManager().getNbDetailTextureFiles(); i++) {
			pr.println(terrain.getManager().getDetailsFiles()[i]);
		}
		for (int j = 0; j < terrain.getTilesZ()/16; j++) {
			for (int i = 0; i < terrain.getTilesX()/16; i++) {
				pr.print(terrain.getManager().getDetailIndexes()[i][j]+" ");
			}
			pr.println();
		}
		Map<String,String> decalTextureFiles = new HashMap<String,String>();
		int count = 0;
		for (Iterator<Decal> it = terrain.getManager().getDecals().values().iterator(); it.hasNext();) {
			Decal decal = it.next();
			String fileName = getMediaRelativePath(decal.getTexture().getImageLocation());
			if (decalTextureFiles.get(fileName) == null) {
				decalTextureFiles.put(fileName,Integer.toString(count));
				count++;
			}
		}
		pr.println(decalTextureFiles.size());
		for (Iterator<String> it = decalTextureFiles.keySet().iterator(); it.hasNext();) {
			pr.println(it.next());
		}
		pr.println(terrain.getManager().getDecals().size());
		for (Iterator<Decal> it = terrain.getManager().getDecals().values().iterator(); it.hasNext();) {
			Decal decal = it.next();
			pr.println(decal.getName());
			pr.println(decal.getUp().x+","+decal.getUp().y+","+decal.getUp().z);
			pr.println(decal.getProjectorPos().x+","+decal.getProjectorPos().y+","+decal.getProjectorPos().z);
			pr.println(decalTextureFiles.get(getMediaRelativePath(decal.getTexture().getImageLocation())));
			pr.println(decal.getSize());
			pr.println(decal.getNear());
			pr.println(decal.getFar());
			pr.println(decal.getFov());
			pr.println(decal.getAspect());
		}
		pr.close();
	}
	
	private static String getMediaRelativePath(String absolutePath) {
		return (absolutePath.substring(absolutePath.indexOf("media")));
	}

	public File[] getDecalsFiles() {
		return decalsFiles;
	}

	public int[][] getDetailIndexes() {
		return detailIndexes;
	}

	public File[] getDetailsFiles() {
		return detailsFiles;
	}

	public File getHeightmapFile() {
		return heightmapFile;
	}

	public int getNbDecalFiles() {
		return nbDecalFiles;
	}

	public int getNbDecals() {
		return nbDecals;
	}

	public int getNbDetailTextureFiles() {
		return nbDetailTextureFiles;
	}

	public File getStampingFile() {
		return stampingFile;
	}

	public HashMap<String, Decal> getDecals() {
		return decals;
	}

	public int getTileRealSize() {
		return tileRealSize;
	}

	public int getTilesX() {
		return tilesX;
	}

	public int getTilesZ() {
		return tilesZ;
	}

	public short[][] getInpassMap() {
		return inpassMap;
	}
}
